import json
from pathlib import Path

from factorio_randovania_mod import creator


def test_create_no_game(tmp_path: Path, test_files: Path) -> None:
    with test_files.joinpath("patcher_a.json").open() as f:
        patch_data = json.load(f)

    creator.create(None, patch_data, tmp_path.joinpath("mod"))

    files = list(tmp_path.joinpath("mod").rglob("*"))
    assert len(files) == 22



from PyInstaller.utils.hooks import collect_data_files

# https://pyinstaller.readthedocs.io/en/stable/hooks.html#provide-hooks-with-package

datas = collect_data_files("factorio_randovania_mod", excludes=["__pyinstaller"])



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import PIL.Image

if TYPE_CHECKING:
    from pathlib import Path

# Code copied from https://stackoverflow.com/a/7274986
# It's licensed under CC BY-SA 3.0


def rgb_to_hsv(rgb: np.ndarray) -> np.ndarray:
    # Translated from source of colorsys.rgb_to_hsv
    # r,g,b should be a numpy arrays with values between 0 and 255
    # rgb_to_hsv returns an array of floats between 0.0 and 1.0.
    rgb = rgb.astype("float")
    hsv = np.zeros_like(rgb)
    # in case an RGBA array was passed, just copy the A channel
    hsv[..., 3:] = rgb[..., 3:]
    r, g, b = rgb[..., 0], rgb[..., 1], rgb[..., 2]
    maxc = np.max(rgb[..., :3], axis=-1)
    minc = np.min(rgb[..., :3], axis=-1)
    hsv[..., 2] = maxc
    mask = maxc != minc
    hsv[mask, 1] = (maxc - minc)[mask] / maxc[mask]
    rc = np.zeros_like(r)
    gc = np.zeros_like(g)
    bc = np.zeros_like(b)
    rc[mask] = (maxc - r)[mask] / (maxc - minc)[mask]
    gc[mask] = (maxc - g)[mask] / (maxc - minc)[mask]
    bc[mask] = (maxc - b)[mask] / (maxc - minc)[mask]
    hsv[..., 0] = np.select([r == maxc, g == maxc], [bc - gc, 2.0 + rc - bc], default=4.0 + gc - rc)
    hsv[..., 0] = (hsv[..., 0] / 6.0) % 1.0
    return hsv


def hsv_to_rgb(hsv: np.ndarray) -> np.ndarray:
    # Translated from source of colorsys.hsv_to_rgb
    # h,s should be a numpy arrays with values between 0.0 and 1.0
    # v should be a numpy array with values between 0.0 and 255.0
    # hsv_to_rgb returns an array of uints between 0 and 255.
    rgb = np.empty_like(hsv)
    rgb[..., 3:] = hsv[..., 3:]
    h, s, v = hsv[..., 0], hsv[..., 1], hsv[..., 2]
    i = (h * 6.0).astype("uint8")
    f = (h * 6.0) - i
    p = v * (1.0 - s)
    q = v * (1.0 - s * f)
    t = v * (1.0 - s * (1.0 - f))
    i = i % 6
    conditions = [s == 0.0, i == 1, i == 2, i == 3, i == 4, i == 5]
    rgb[..., 0] = np.select(conditions, [v, q, p, p, t, v], default=v)
    rgb[..., 1] = np.select(conditions, [v, v, v, q, p, p], default=t)
    rgb[..., 2] = np.select(conditions, [v, p, t, v, v, q], default=p)
    return rgb.astype("uint8")


def shift_hue(img: PIL.Image.Image, hue: float) -> PIL.Image.Image:
    arr = np.array(img.convert("RGBA"))

    hsv = rgb_to_hsv(arr)
    hsv[..., 0] = hue
    rgb = hsv_to_rgb(hsv)

    return PIL.Image.fromarray(rgb, "RGBA")


def hue_shift(input_path: Path, output_path: Path, rotation: float) -> None:
    """Creates a new image, by shifting the hue of the input image."""
    img = PIL.Image.open(input_path)
    new_img = shift_hue(img, rotation / 360.0)
    output_path.parent.mkdir(parents=True, exist_ok=True)
    new_img.save(output_path)



from __future__ import annotations

import json
import typing
from pathlib import Path

from jsonschema import Draft7Validator

if typing.TYPE_CHECKING:
    from factorio_randovania_mod.configuration import Configuration

_SCHEMA_PATH = Path(__file__).parent.joinpath("files", "schema.json")


def _read_schema() -> dict:
    with _SCHEMA_PATH.open() as f:
        return json.load(f)


def validate(data: dict) -> Configuration:
    Draft7Validator(_read_schema()).validate(data)

    return typing.cast("Configuration", data)



import numpy as np
import pytest
from factorio_randovania_mod import color_util


@pytest.mark.parametrize(
    ("rgb", "expected"),
    [
        ((255, 0, 0), (0, 1, 255)),
        ((0, 255, 0), (1 / 3, 1, 255)),
        ((0, 0, 255), (2 / 3, 1, 255)),
        ((255, 255, 255), (0, 0, 255)),
        ((0, 0, 0), (0, 0, 0)),
    ],
)
def test_rgb_to_hsv(rgb, expected):
    arr = np.array(rgb)
    result = color_util.rgb_to_hsv(arr)
    assert (result == np.array(expected)).all()



from __future__ import annotations

import string
import typing


def wrap_array_pretty(data: list) -> str:
    return "{\n    " + ",\n    ".join(wrap(item, "    ") for item in data) + "\n}"


def _dict_key(key: str) -> str:
    if set(key).issubset(string.ascii_letters + "_"):
        return key
    else:
        return f'["{key}"]'


def wrap(data: typing.Any, indent: str = "") -> str:
    if isinstance(data, list | tuple):
        return "{" + ", ".join(wrap(item, indent) for item in data) + "}"

    if isinstance(data, dict):
        return (
            "{\n"
            + "\n".join(
                f"{indent}    {_dict_key(key)} = {wrap(value, f'{indent}    ')}," for key, value in data.items()
            )
            + f"\n{indent}}}"
        )

    if isinstance(data, bool):
        return "true" if data else "false"

    if data is None:
        return "nil"

    if isinstance(data, str):
        return f'"{data}"'

    return str(data)



import typing

from factorio_randovania_mod.configuration import ConfigurationRecipesItem


class TechCosts(typing.TypedDict):
    count: int
    time: int
    ingredients: list[tuple[str, int]]


class CustomTechTreeItem(typing.TypedDict):
    name: str
    icon_size: typing.NotRequired[int]
    icon: str
    costs: TechCosts
    prerequisites: list[str] | None
    take_effects_from: typing.NotRequired[str]


class TechRepurpose(typing.TypedDict):
    science_pack: str
    prerequisites: list[str]


class GeneratedFiles(typing.TypedDict):
    tech_tree: list[CustomTechTreeItem]
    local_unlocks: dict[str, list[str]]
    existing_tree_repurpose: dict[str, TechRepurpose]
    starting_tech: list[str]
    custom_recipes: list[ConfigurationRecipesItem]



import argparse
import json
import logging
import logging.config
import time
from pathlib import Path

from factorio_randovania_mod import creator


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--factorio-path",
        required=True,
        type=Path,
        help="Path to where a Factorio installation can be found.",
    )
    parser.add_argument(
        "--output-path",
        required=True,
        type=Path,
        help="Path to where the mod files will be written to.",
    )
    parser.add_argument("--input-json", required=True, type=Path, help="Path to the configuration json.")
    parser.add_argument("-q", "--quiet", action="store_true", help="Disables all info and debug logs.")
    return parser


def setup_logging() -> None:
    handlers = {
        "default": {
            "level": "DEBUG",
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",  # Default is stderr
        },
    }
    logging.config.dictConfig(
        {
            "version": 1,
            "formatters": {
                "default": {
                    "format": "[%(asctime)s] [%(levelname)s] [%(name)s] %(funcName)s: %(message)s",
                }
            },
            "handlers": handlers,
            "disable_existing_loggers": False,
            "loggers": {
                "default": {
                    "level": "DEBUG",
                },
            },
            "root": {
                "level": "DEBUG",
                "handlers": list(handlers.keys()),
            },
        }
    )


def main() -> None:
    setup_logging()
    parser = create_parser()
    args = parser.parse_args()
    if args.quiet:
        logging.getLogger().setLevel(logging.WARNING)

    with args.input_json.open() as f:
        configuration = json.load(f)

    start = time.time()
    creator.create(
        args.factorio_path,
        configuration,
        args.output_path,
    )
    end = time.time()
    print(f"Patcher took {end - start:.03f} seconds")



# This file is generated. Manual changes will be lost
# fmt: off
# ruff: noqa
from __future__ import annotations

import typing_extensions as typ




# Schema entries
@typ.final
class ConfigurationTechnologiesItemCost(typ.TypedDict):
    count: int
    time: int
    ingredients: list[str]

@typ.final
class ConfigurationTechnologiesItem(typ.TypedDict):
    tech_name: typ.NotRequired[str]
    locale_name: str
    description: str
    icon: str
    icon_size: typ.NotRequired[int]
    cost: ConfigurationTechnologiesItemCost
    prerequisites: list[str]
    unlocks: list[str]

@typ.final
class ConfigurationRecipesItemIngredientsItem(typ.TypedDict):
    name: str
    amount: int
    type: typ.NotRequired[str]

@typ.final
class ConfigurationRecipesItem(typ.TypedDict):
    recipe_name: str
    category: str
    result_amount: int
    ingredients: list[ConfigurationRecipesItemIngredientsItem]

@typ.final
class Configuration(typ.TypedDict):
    configuration_identifier: str
    layout_uuid: str
    technologies: list[ConfigurationTechnologiesItem]
    recipes: list[ConfigurationRecipesItem]
    starting_tech: list[str]



from __future__ import annotations

import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import configparser
    from pathlib import Path


def get_from_locale(locale: configparser.ConfigParser, group: str, n: str) -> str:
    if n in locale[group]:
        return locale[group][n]
    if f"{n}-1" in locale[group]:
        return locale[group][f"{n}-1"]

    i = n.rfind("-")
    if i != -1:
        front, number = n[:i], n[i + 1 :]
        if number.isdigit():
            return get_from_locale(locale, group, front)

    raise KeyError(n)


def get_localized_name(locale: configparser.ConfigParser, n: str) -> str:
    for k in [
        "item-name",
        "entity-name",
        "fluid-name",
        "equipment-name",
        "recipe-name",
        "technology-name",
    ]:
        if n in locale[k]:
            return locale[k][n]
        if f"{n}-1" in locale[k]:
            return locale[k][f"{n}-1"]

    if n.startswith("fill-"):
        return f"Fill {locale['fluid-name'][n[5:-7]]} barrel"

    if n.endswith("-barrel"):
        return f"{locale['fluid-name'][n[:-7]]} barrel"

    hardcoded_names = {
        "solid-fuel-from-heavy-oil": "Solid Fuel (Heavy Oil)",
        "solid-fuel-from-light-oil": "Solid Fuel (Light Oil)",
        "solid-fuel-from-petroleum-gas": "Solid Fuel (Petroleum Gas)",
    }

    try:
        return hardcoded_names[n]
    except KeyError:
        i = n.rfind("-")
        if i != -1:
            front, number = n[:i], n[i + 1 :]
            if number.isdigit():
                return f"{get_localized_name(locale, front)} {number}"
        raise


def ensure_locale_read(locale: configparser.ConfigParser, files: list[Path]) -> None:
    filenames = [os.fspath(filename) for filename in files]
    did_read = locale.read(filenames)
    if did_read != filenames:
        missing = set(filenames) - set(did_read)
        raise FileNotFoundError(str(missing))



from __future__ import annotations

import collections
import configparser
import shutil
import typing
from pathlib import Path

from factorio_randovania_mod import schema
from factorio_randovania_mod.color_util import hue_shift
from factorio_randovania_mod.locale_lib import ensure_locale_read, get_from_locale
from factorio_randovania_mod.lua_util import wrap, wrap_array_pretty

if typing.TYPE_CHECKING:
    from factorio_randovania_mod.configuration import (
        ConfigurationTechnologiesItem,
    )
    from factorio_randovania_mod.mod_lua_api import CustomTechTreeItem, GeneratedFiles

_TEMPLATE_PATH = Path(__file__).parent.joinpath("lua_src")


def create_hue_shifted_images(factorio_path: Path, output_path: Path) -> None:
    lab_angle = 120.0
    assembler_angle = 310.0

    base_graphics_path = factorio_path.joinpath("data/base/graphics")

    lab_images = [
        "entity/lab/lab.png",
        "entity/lab/lab-light.png",
        "entity/lab/hr-lab.png",
        "entity/lab/hr-lab-light.png",
        "icons/lab.png",
    ]
    assembler_images = [
        "entity/assembling-machine-1/assembling-machine-1.png",
        "entity/assembling-machine-1/hr-assembling-machine-1.png",
        "icons/assembling-machine-1.png",
    ]

    conversions = []

    for path in lab_images:
        conversions.append(
            (
                path,
                path.replace("lab", "burner-lab"),
                lab_angle / 360.0,
            )
        )

    for path in assembler_images:
        conversions.append(
            (
                path,
                path.replace("assembling-machine-1", "burner-assembling-machine"),
                assembler_angle / 360.0,
            )
        )

    for source_path, target_path, rotation in conversions:
        hue_shift(
            base_graphics_path.joinpath(source_path),
            output_path.joinpath("graphics", target_path),
            rotation,
        )


def process_technology(
    output_locale: configparser.ConfigParser,
    local_unlocks: dict[str, list[str]],
    progressive_sources: dict[tuple[str, ...], list[str]],
    tech: ConfigurationTechnologiesItem,
    source_locale: configparser.ConfigParser | None,
) -> CustomTechTreeItem:
    """
    Process an entry of patch_data["technologies"]
    :param output_locale:
    :param local_unlocks:
    :param progressive_sources:
    :param tech:
    :param source_locale:
    :return: A new entry for tech-tree.lua
    """
    tech_name = tech["tech_name"]
    output_locale["technology-name"][tech_name] = tech["locale_name"]
    output_locale["technology-description"][tech_name] = tech["description"]

    new_tech: CustomTechTreeItem = {
        "name": tech_name,
        "icon": tech["icon"],
        "costs": {
            "count": tech["cost"]["count"],
            "time": tech["cost"]["time"],
            "ingredients": [(it, 1) for it in tech["cost"]["ingredients"]],
        },
        "prerequisites": tech["prerequisites"] if tech["prerequisites"] else None,
        # "fake_effects": tech["fake_effects"],
    }
    if "icon_size" in tech:
        new_tech["icon_size"] = tech["icon_size"]

    if len(tech["unlocks"]) == 1:
        new_tech["take_effects_from"] = tech["unlocks"][0]
        if source_locale is not None:
            output_locale["technology-description"][tech_name] = get_from_locale(
                source_locale, "technology-description", tech["unlocks"][0]
            )
    elif tech["unlocks"]:
        local_unlocks[tech_name] = tech["unlocks"]
        progressive_sources[tuple(tech["unlocks"])].append(tech_name)

    return new_tech


def generate_output(
    output_path: Path,
    generated_files: GeneratedFiles,
    locale: configparser.ConfigParser,
    factorio_path: Path | None,
) -> None:
    """
    Generates all files for the mod.
    :param output_path: Where to place the output
    :param generated_files: Data for generating all lua files
    :param locale: Used as template
    :param factorio_path: Source for hue shifting the images
    :return:
    """
    shutil.copytree(_TEMPLATE_PATH, output_path)
    output_path.joinpath("generated").mkdir()

    def generate_file(name: str, content: str) -> None:
        output_path.joinpath("generated", name).write_text("return " + content)

    generate_file("tech-tree.lua", wrap_array_pretty(generated_files["tech_tree"]))
    generate_file("local-unlocks.lua", wrap(generated_files["local_unlocks"]))
    generate_file("existing-tree-repurpose.lua", wrap(generated_files["existing_tree_repurpose"]))

    generate_file("starting-tech.lua", wrap_array_pretty(generated_files["starting_tech"]))
    generate_file("custom-recipes.lua", wrap_array_pretty(generated_files["custom_recipes"]))

    with output_path.joinpath("locale/en/strings.cfg").open("w") as f:
        locale.write(f, space_around_delimiters=False)

    if factorio_path is not None:
        create_hue_shifted_images(factorio_path, output_path)


def create(factorio_path: Path | None, patch_data: dict, output_folder: Path) -> None:
    output_path = output_folder.joinpath("randovania-layout")
    shutil.rmtree(output_path, ignore_errors=True)

    configuration = schema.validate(patch_data)

    original_locale: configparser.ConfigParser | None = None
    if factorio_path is not None:
        original_locale = configparser.ConfigParser()
        ensure_locale_read(
            original_locale,
            [
                factorio_path.joinpath("data/base/locale/en/base.cfg"),
                _TEMPLATE_PATH.joinpath("locale/en/strings.cfg"),
            ],
        )

    locale = configparser.ConfigParser()
    ensure_locale_read(
        locale,
        [
            _TEMPLATE_PATH.joinpath("locale/en/strings.cfg"),
        ],
    )

    progressive_sources: dict[tuple[str, ...], list[str]] = collections.defaultdict(list)
    generated_files: GeneratedFiles = {
        "tech_tree": [],
        "local_unlocks": {},
        "existing_tree_repurpose": {},
        "starting_tech": configuration["starting_tech"],
        "custom_recipes": configuration["recipes"],
    }

    for tech in configuration["technologies"]:
        generated_files["tech_tree"].append(
            process_technology(
                locale,
                generated_files["local_unlocks"],
                progressive_sources,
                tech,
                original_locale,
            )
        )

    # TODO: add the offworld research to `existing_tree_repurpose`

    for progressive_sequence, sources in progressive_sources.items():
        for i, tech_name in enumerate(progressive_sequence):
            if i == 0:
                prerequisites = sources
            else:
                prerequisites = [progressive_sequence[i - 1]]

            generated_files["existing_tree_repurpose"][tech_name] = {
                "science_pack": "impossible-science-pack",
                "prerequisites": prerequisites,
            }

    generate_output(output_path, generated_files, locale, factorio_path)
